#!/usr/bin/env python
# coding: utf-8

# # Introduction
# 
# Many imaging studies require T1-weighted images with specific acquisition parameters, usually sagittal acquisition. In PPMI, metadata on MRI acquisition is available in file "Magnetic_Resonance_Imaging__MRI_.csv", however, this file does not contain detailed information about acquistion parameters.
# 
# This notebook downloads acquisition parameters of 3D T1-weighted images from the PPMI imaging database, filters sagittal acquisition scans, and converts visit names used in the imaging database to the ones used in other PPMI metadata. 
# 
# The resulting file can be used to build imaging cohorts based on PPMI.

# In[1]:


from IPython.display import HTML

HTML('''<script>
code_show=true; 
function code_toggle() {
 if (code_show){
 $('div.input').hide();
 } else {
 $('div.input').show();
 }
 code_show = !code_show
} 
$( document ).ready(code_toggle);
</script>
<form action="javascript:code_toggle()"><input type="submit" value="Click here to toggle on/off the Python code."></form>''')


# # Data download
# 
# Let's download information about PPMI 3D T1-weighted scans:

# In[2]:


import os
import os.path as op
import pandas as pd
import ppmi_downloader

data_dir = 'data'

if not op.exists(data_dir):
    os.makedirs(data_dir)

mri_file_name = '3D_mri_info.csv'
required_files = [mri_file_name]
missing_files = [x for x in required_files if not op.exists(os.path.join(data_dir, x))]

if len(missing_files) > 0:
    ppmi = ppmi_downloader.PPMIDownloader()
    file_name = ppmi.download_3D_T1_info(destination_dir=data_dir, headless=False)
    assert(op.exists(op.join(data_dir, file_name)))
    os.rename(op.join(data_dir, file_name), op.join(data_dir, mri_file_name))

print('File downloaded')


# # Filter sagittal acquisitions
# 
# PPMI scans were acquired using the following protocols:

# In[3]:


pd.set_option('display.max_rows', 500)
mri_info = pd.read_csv(op.join(data_dir, mri_file_name))
mri_info.groupby('Description').count()


# To keep only the sagittal acquisitions, we will remove the following protocols:

# In[4]:


# Remove sequences that exactly match the following
removed_sequences=['COR', # coronal acquisitions
                   'Coronal', 
                   'Cal Head 24',  # not sure what this is
                   'Transverse',  # transverse (axial) acquisitions
                   'tra_T1_MPRAGE',
                   'TRA'
                  ]
print(removed_sequences)


# We will also remove the protocol names containing the following strings:

# In[5]:


# Remove sequences containing the following strings
removed_sequences_contain = ['AX', 'axial', 'Phantom']
print(removed_sequences_contain)


# We obtain the following list of protocols:

# In[6]:


mri_info = mri_info[~mri_info['Description'].isin(removed_sequences)]
for s in removed_sequences_contain:
    mri_info = mri_info[~mri_info['Description'].str.contains(s)]
mri_info.groupby('Description').count()


# # Convert visit names

# Let's use the following map to convert visit names to the codes used in PPMI metadata:

# In[7]:


visit_map = {
    'Screening': 'SC',
    'Baseline': 'BL',
    'Month 12': 'V04',
    'Month 24': 'V06',
    'Month 36': 'V08',
    'Month 48': 'V10',
    'Symptomatic Therapy': 'ST',
    'Unscheduled Visit 01': 'U01',
    'Unscheduled Visit 02': 'U02',
    'Premature Withdrawal': 'PW'
}
print(visit_map)


# We obtain the following distribution by visit code:

# In[8]:


mri_info['Visit code'] = mri_info['Visit'].apply(lambda x: visit_map[x])
mri_info.groupby('Visit code').count()


# Finally, let's save our table as csv file:

# In[9]:


filename = 'MRI_info.csv'
mri_info.to_csv(op.join(data_dir, filename), index=False)
print(f'Saved in {filename}')


# In[ ]:







"""Generate the `run.py` file.

It creates convinience methods to evoke the scripts generated from notebooks.
"""
import pathlib
import os
from string import Template


notebooks_dir = pathlib.Path(__file__).parents[1].joinpath("notebooks").resolve()
scripts_dir = pathlib.Path(__file__).parents[1].joinpath("scripts").resolve()
output_file = pathlib.Path(__file__).parent.joinpath("run.py").resolve()

with open(output_file, "w") as fout:
    fout.write(
        """\"\"\"Convinience methods to run the auto-generated scripts.

Notes
-----
The methods below need to be created manually when a new notebook is added.
\"\"\"
import importlib
"""
    )

for root, dirs, files in os.walk(notebooks_dir):
    if root.endswith(".ipynb_checkpoints"):
        continue

    for filename in files:
        rel_path = pathlib.Path(root).joinpath(filename).relative_to(notebooks_dir)
        script_path = scripts_dir.joinpath(str(rel_path).removesuffix(".ipynb") + ".py")

        if os.path.exists(script_path):
            with open(script_path) as fin, open(output_file, "a") as fout:
                s = Template(
                    """

def $func_name():
    \"\"\"Execute auto-generated script for `../notebooks/${func_name}.ipynb`.\"\"\"
    importlib.import_module(\"$func_name\", \"livingpatk_utils.scripts\")
"""
                )
                fout.write(
                    s.safe_substitute(
                        {
                            "func_name": str(rel_path).removesuffix(".ipynb"),
                        }
                    )
                )



"""Convinience methods to run the auto-generated scripts.

Notes
-----
The methods below need to be created manually when a new notebook is added.
"""
import importlib


def mri_metadata():
    """Execute auto-generated script for `../notebooks/mri_metadata.ipynb`."""
    importlib.import_module("mri_metadata", "livingpatk_utils.scripts")


def pd_status():
    """Execute auto-generated script for `../notebooks/pd_status.ipynb`."""
    importlib.import_module("pd_status", "livingpatk_utils.scripts")



#!/usr/bin/env python
# coding: utf-8

# # 👋 Introduction

# The medication status of PD patients is important as medication importantly affects clinical measures such as the Hoehn & Yahr score used in many studies. In PPMI, medication is available through the following main variables:
# * PDSTATE (ON/OFF): the current functional state of the patient
# * PDTRTMNT (0/1): 1 if the participant is on PD medication or receives deep brain stimulation, 0 otherwise
# * PDMEDTM: time of most recent PD medication dose
# * PDMEDDT: date of most recent PD medication dose
# 
# As mentioned in the "Methods for Defining PD Med Use" in PPMI study data, OFF state requires that the last dose of levodopa or dopamine agonist was taken 6 hours or more before MDS-UPDRS Part III assessment.
# 
# The goal of this notebook is (1) to identify and correct inconsistencies among these variables, (2) to impute missing data for PDSTATE and PDTRTMNT, and (3) to check the sanity of the corrected dataset.

# In[1]:


from IPython.display import HTML

HTML(
    """<script>
code_show=true; 
function code_toggle() {
 if (code_show){
 $('div.input').hide();
 } else {
 $('div.input').show();
 }
 code_show = !code_show
} 
$( document ).ready(code_toggle);
</script>
<form action="javascript:code_toggle()"><input type="submit" value="Click here to toggle on/off the Python code."></form>"""
)


# In[2]:


import datetime
import warnings

import pytz


warnings.filterwarnings('ignore')

now = datetime.datetime.now(pytz.utc).strftime('%Y-%m-%d %H:%M:%S %Z %z')
print(f"This notebook was run on {now}")


# # 🔽 Data download
# 
# The above-mentioned variables are available in PPMI file `MDS_UPDRS_Part_III.csv`. To download this file, we will use package `ppmi-downloader` available on PyPi. The package will ask you for your PPMI login and password.

# In[3]:


import os
import os.path as op

import pandas as pd
import ppmi_downloader


data_dir = "data"

if not op.exists(data_dir):
    os.makedirs(data_dir)
required_files = ["MDS_UPDRS_Part_III.csv"]
missing_files = [x for x in required_files if not op.exists(os.path.join(data_dir, x))]

if len(missing_files) > 0:
    ppmi = ppmi_downloader.PPMIDownloader()
    ppmi.download_metadata(
        missing_files, destination_dir=data_dir, headless=False, timeout=600
    )

df = pd.read_csv(op.join(data_dir, "MDS_UPDRS_Part_III.csv"))

print("File downloaded")


# # ⁉️ Inconsistencies

# ## Records with no value

# <div class="alert alert-block alert-danger">
#      	&#10060; <b>Problem:</b> some records have missing data for all UPDRS-III variables.
# </div>

# Number of records with NaNs for all UPDRS-III variables:

# In[4]:


updrs3_vars = [
    "PDMEDDT",
    "PDMEDTM",
    "PDSTATE",
    "EXAMTM",
    "DBS_STATUS",
    "NP3SPCH",
    "NP3FACXP",
    "NP3RIGN",
    "NP3RIGRU",
    "NP3RIGLU",
    "NP3RIGRL",
    "NP3RIGLL",
    "NP3FTAPR",
    "NP3FTAPL",
    "NP3HMOVR",
    "NP3HMOVL",
    "NP3PRSPR",
    "NP3PRSPL",
    "NP3TTAPR",
    "NP3TTAPL",
    "NP3LGAGR",
    "NP3LGAGL",
    "NP3RISNG",
    "NP3GAIT",
    "NP3FRZGT",
    "NP3PSTBL",
    "NP3POSTR",
    "NP3BRADY",
    "NP3PTRMR",
    "NP3PTRML",
    "NP3KTRMR",
    "NP3KTRML",
    "NP3RTARU",
    "NP3RTALU",
    "NP3RTARL",
    "NP3RTALL",
    "NP3RTALJ",
    "NP3RTCON",
    "NP3TOT",
    "DYSKPRES",
    "DYSKIRAT",
    "NHY",
    "DBSONTM",
    "DBSOFFTM",
    "HRPOSTMED",
    "HRDBSOFF",
    "HRDBSON",
]


# In[5]:


len(df[df[updrs3_vars].isnull().all(axis=1)])


# <div class="alert alert-block alert-success">
#      	&#10003; <b>Proposed correction</b>: remove these records.
# </div>

# ⚙️ Implementation
# 
# 

# In[6]:


df = df[df[updrs3_vars].notna().any(axis=1)]


# Updated number of records with NaNs for all UPDRS-III variables:

# In[7]:


len(df[df[updrs3_vars].isnull().all(axis=1)])


# ## PDTRTMNT=0 and PDSTATE=ON
# 
# <div class="alert alert-block alert-danger">
#      	&#10060; <b>Problem:</b> a few records have PDSTATE=ON and PDTRTMNT=0, which is inconsistent:
# </div>

# In[8]:


df.groupby(["PDSTATE", "PDTRTMNT"], dropna=False)[["REC_ID"]].count()


# The lines below show the difference between EXAMTM and PDMEDTM. All the records have a PDMEDTM that is earlier to EXAMTM by at most 5 hours:

# In[9]:


errors = df[(df["PDSTATE"] == "ON") & (df["PDTRTMNT"] == 0)]
# print the time difference between EXAMTM and PDMEDTM
(pd.to_datetime(errors["EXAMTM"]) - pd.to_datetime(errors["PDMEDTM"]))


# <div class="alert alert-block alert-success">
#      	&#10003; <b>Proposed correction</b>: set PDTRTMNT to 1 for these records. It doesn't seem realistic that a plausible PDMEDTM and PDSTATE=ON have been entered by mistake while the patient was not under medication.
# </div>

# ⚙️ Implementation
# 
# 

# In[10]:


df.loc[(df["PDSTATE"] == "ON") & (df["PDTRTMNT"] == 0), "PDTRTMNT"] = 1


# Let's verify that the inconsistency is now resolved:

# In[11]:


df.groupby(["PDSTATE", "PDTRTMNT"], dropna=False)[["REC_ID"]].count()


# ## PDTRTMNT=0 and PDMEDTM not empty
# 
# <div class="alert alert-block alert-danger">
#  	&#10060; <b>Problem:</b> some records have a non-empty PDMEDTIME and have PDTRTMNT=0, which is inconsistent.
#     </div>

# Number of records with a non-empty PDMEDTIME and PDTRTMNT=0:

# In[12]:


errors = df[(df["PDMEDTM"].notnull()) & (df["PDTRTMNT"] == 0)]
len(errors)


# PDMEDTIME values for these records look plausible:

# In[13]:


from matplotlib import pyplot as plt

pd.to_datetime(errors["PDMEDTM"]).dt.hour.hist(bins=24, xrot=90, legend=True)
plt.xlabel("Hour of the day")
plt.ylabel("Number of records")
plt.show()


# <div class="alert alert-block alert-success">
#     	&#10003; <b>Proposed correction:</b> set PDTRTMNT=1. It is unlikely that a plausible medication time was entered by mistake.
# </div>

# ⚙️ Implementation

# In[14]:


df.loc[(df["PDTRTMNT"] == 0) & (df["PDMEDTM"].notnull()), "PDTRTMNT"] = 1


# Let's verify that the inconsistency is now fixed by counting the number of records with a non-empty PDMEDTIME and PDTRTMNT=0:

# In[15]:


errors = df[(df["PDMEDTM"].notnull()) & (df["PDTRTMNT"] == 0)]
len(errors)


# ## EVENT_ID = SC and PDTRTMNT = 1

# <div class="alert alert-block alert-danger">
#      	&#10060; <b>Problem</b>: Some patients were on medication at screening time while PPMI patients were supposed to be unmedicated at screening time.
#     </div>

# Number of patients on medication at screening time:

# In[16]:


len(df[(df["EVENT_ID"] == "SC") & (df["PDTRTMNT"] == 1)])


# <div class="alert alert-block alert-success">
#      	&#10003; <b>Proposed correction:</b> keep the records. Maybe the patients started medication between recruitment and screening time.
# </div>

# ## PDSTATE=ON and EXAMTM<PDMEDTM

# <div class="alert alert-block alert-danger">
#      	&#10060; <b>Problem</b>: Some records have PDSTATE=ON but PDMEDTM is after EXAMTM. 
#     </div>

# Number of records where PDSTATE=ON and EXAMTM<PDMEDTM:

# In[17]:


# ON records
on = df[df["PDSTATE"] == "ON"]


def to_secs(x):
    """
    Convert time from hh:mm to seconds since midnight

    x: time in hh:ss format
    return: number of seconds elapsed since midnight
    """
    if str(x) == "nan":
        import numpy as np

        return np.NaN
    try:
        hour, mn, sec = x.split(":")
    except Exception as e:
        print(f'Cannot process "{x}"')
        raise (e)
    return int(hour) * 3600 + int(mn) * 60 + int(sec)


on["delta"] = on["EXAMTM"].apply(to_secs) - on["PDMEDTM"].apply(to_secs)
len(on[on["delta"] < 0])


# <div class="alert alert-block alert-success">
#      	&#10003; <b>Proposed correction:</b> discard the records. 
# </div>

# Note: for some of these records, medication date was likely on the day before the exam, although this cannot be verified since dates only contain a year and a month but no day. Even in such cases, duration between medication time and exam time was more than 6 hours which is inconsistent with the rule used for the other records.

# ⚙️ Implementation

# In[18]:


before = len(df)
df = df[
    ~(
        (df["PDSTATE"] == "ON")
        & df["EXAMTM"].notnull()
        & df["PDMEDTM"].notnull()
        & (df["EXAMTM"] < df["PDMEDTM"])
    )
]
print(f"Removed {before-len(df)} records where PDSTATE=ON and EXAMTM<PDMEDTM")


# ## Visits with 3 exams

# <div class="alert alert-block alert-danger">
#     	&#10060; <b>Problem</b>: some visits have 3 exams while a maximum of two exams per visit are expected, one in OFF state and one in ON state.
# </div>

# Number of records that belong to a visit with more than 3 exams:

# In[19]:


pb = df.groupby(["PATNO", "EVENT_ID"]).filter(lambda x: len(x) > 2)
len(pb)


# Each exams triple has an exam with missing EXAMTM and missing PDSTATE:

# In[20]:


pb = df.groupby(["PATNO", "EVENT_ID"]).filter(lambda x: len(x) > 2)
pb_trunc = pb[["EVENT_ID", "PDSTATE", "EXAMTM"]]
from IPython.display import HTML

HTML(pb_trunc.to_html(index=False))


# <div class="alert alert-block alert-success">
#   	&#10003;   <b>Proposed correction:</b> remove exam with EXAMTM=NaN and PDSTATE=NaN when visit has 3 exams.
# </div>

# ⚙️ Implementation

# In[21]:


a = df.groupby(["PATNO", "EVENT_ID"]).filter(lambda x: len(x) > 2)
index = (a[(a["PDSTATE"].isnull()) & (a["EXAMTM"].isnull())]).index

before_len = len(df)
df = df[~df.index.isin(index)]
print(f"Number of removed records: {before_len-len(df)}")


# Let's verify that the inconsistency is solved by counting the number of records that belong to a visit with more than 3 exams:

# In[22]:


pb = df.groupby(["PATNO", "EVENT_ID"]).filter(lambda x: len(x) > 2)
len(pb)


# # 🕵️ Imputation of missing PDSTATE or PDTRTMNT

# <div class="alert alert-block alert-danger">
#     	&#10060; <b>Problem:</b> variables PDSTATE and PDTRTMNT have missing data, which makes it difficult to identify when/if a patient was under medication.
# </div>

# The following table summarizes the number of records for which PDSTATE or PDTRTMNT is missing:

# In[23]:


df.groupby(["PDSTATE", "PDTRTMNT"], dropna=False)[["REC_ID"]].count()


# The following cases will be treated separately in the following sections:
# 
# |               |PDSTATE | PDTRTMNT | Number of records |
# |---------------|--------|----------|-------------------|
# | **Case 1**  | OFF    | NaN      | 6                 |
# | **Case 2** | NaN    | 0        | 6947          |
# | **Case 3**|  NaN   | 1        | 633               |
# | **Case 4** | NaN    | NaN      | 2322             |
# | **Case 5**  | ON     | NaN      | 1                 |

# ## Case 1: PDSTATE=OFF and PDTRTMNT=NaN

# No record in case 1 has a medication date (PDMEDDT), a medication time (PDMEDTM), or
# a DBS status (DBS_STATUS):

# In[24]:


case_1 = df[(df["PDSTATE"] == "OFF") & (df["PDTRTMNT"].isnull())]
case_1.groupby(
    [
        "PDMEDDT",
        "PDMEDTM",
        "DBS_STATUS",
        "HRPOSTMED",
        "DBSONTM",
        "DBSOFFTM",
        "HRDBSOFF",
        "HRDBSON",
    ],
    dropna=False,
)[["REC_ID"]].count()


# <div class="alert alert-block alert-success">
#  	&#10003;    <b>Proposed correction</b>: set PDTRTMNT=0. It is unlikely that these records correspond to medicated patients when none of the variables related to medication have a value.
# </div>

# ⚙️ Implementation

# In[25]:


df.loc[(df["PDSTATE"] == "OFF") & (df["PDTRTMNT"].isnull()), "PDTRTMNT"] = 0


# Let's verify that case 1 is now resolved:

# In[26]:


df.groupby(["PDSTATE", "PDTRTMNT"], dropna=False)[["REC_ID"]].count()


# ## Case 2: PDSTATE=NaN and PDTRTMNT=0
# 
# <div class="alert alert-block alert-success">
#    	&#10003;   <b>Proposed correction</b>: set PDSTATE=OFF. The patient is not medicated and for this reason PDSTATE is likely to not have been entered.
# </div>

# ⚙️ Implementation

# In[27]:


df.loc[(df["PDSTATE"].isnull()) & (df["PDTRTMNT"] == 0), "PDSTATE"] = "OFF"


# Let's verify that case 2 is now resolved:

# In[28]:


df.groupby(["PDSTATE", "PDTRTMNT"], dropna=False)[["REC_ID"]].count()


# ## Case 3: PDSTATE=NaN and PDTRTMNT=1
# 
# <div class="alert alert-block alert-success">
#     	&#10003; <b>Proposed correction</b>: 
#     <ul>
#         <li>(a) <b>IF</b> record belongs to a visit with two exams:</li>
#             <ul>
#                 <li> Set PDSTATE=OFF for record with earliest EXAMTM</li>
#                 <li> Set PDSTATE=ON for record with latest EXAMTM</li>
#             </ul>
#         <li>(b) <b>ELSE</b>: determine PDSTATE as a function of PDMEDTM and EXAMTM:</li>
#             <ul>
#                 <li>(i) <b>IF</b> PDMEDTM or EXAMTM are missing <b>THEN</b> discard record.</li>
#                 <li> (ii) <b>IF</b> PDMEDTM is earlier than EXAMTM:</li>
#                     <ul>
#                         <li><b>IF</b> EXAMTM-PDMEDTM < 6 hours (PPMI cut-off)<b>THEN</b>:
#                 <ul><li><b>IF</b> EXAMTM-PDMEDTM >= 30 min <b>THEN</b> set PDSTATE=ON.</li>
#                     <li><b>ELSE</b>: discard record.</ul>
#         <li><b>ELSE:</b> set PDSTATE=OFF</li>
#                     </ul>
#                 <li> (iii) <b>IF</b> PDMEDTM is later than EXAMTM <b>THEN</b>:
#                     <UL>
#                         <li>(a) <b>IF</b> PDMEDTM is later than 4pm <b>THEN</b> assume record belongs to the previous day, set PDSTATE=OFF.</li>
#                         <li>(b) <b>ELSE</b>                     
#                     discard record.
#                             </ul></li>
#     </ul>
#     
#     
#   
# </div>

# ⚙️ Implementation
# 
# Let's implement each sub-case separately.

# ### Case 3.a: visits with two exams

# Let's count the number of records in case 3.a:

# In[29]:


case_3 = (df["PDSTATE"].isnull()) & (df["PDTRTMNT"] == 1)
a = df[case_3].groupby(["PATNO", "EVENT_ID"]).filter(lambda x: len(x) == 2)
print(f"Found {len(a)} records in Case 3.a")


# ### Case 3.b.i: visits with one exam, missing EXAMTM or PDMEDTM

# Number of records in case 3.b with missing EXAMTM or missing PDMEDTM:

# In[30]:


len(df[case_3 & ((df["EXAMTM"].isnull()) | (df["PDMEDTM"].isnull()))])


# In[31]:


before_len = len(df)
df.drop(
    df[case_3 & ((df["EXAMTM"].isnull()) | (df["PDMEDTM"].isnull()))].index,
    inplace=True,
)
print(f"Removed {before_len-len(df)} record(s) with missing EXAMTM or PDMEDTM")


# Updated records distribution in case 3:

# In[32]:


df.groupby(["PDSTATE", "PDTRTMNT"], dropna=False)[["REC_ID"]].count()


# In[33]:


case_3 = (df["PDSTATE"].isnull()) & (df["PDTRTMNT"] == 1)


# ### Case 3.b.ii: visits with one exam, PDMEDTM earlier than EXAMTM

# Number of records in case 3.b where PDMEDTM is earlier or equal to EXAMTM:

# In[34]:


len(df[case_3 & (df["PDMEDTM"] <= df["EXAMTM"])])


# In[35]:


def to_secs(x):
    if str(x) == "nan":
        import numpy as np

        return np.NaN
    try:
        hour, mn, sec = x.split(":")
    except Exception as e:
        print(f'Cannot process "{x}"')
        raise (e)
    return int(hour) * 3600 + int(mn) * 60 + int(sec)


# Number of records in case 3.b where PDMEDTM is earlier or equal to EXAMTM and $30\,min \le \text{EXAMTM}-\text{PDMEDTM} < 6\,hours$:

# In[36]:


len(
    df[
        case_3
        & (df["EXAMTM"].apply(to_secs) - df["PDMEDTM"].apply(to_secs) >= 1800)
        & (df["EXAMTM"].apply(to_secs) - df["PDMEDTM"].apply(to_secs) < 6 * 3600)
    ]
)


# Let's set PDSTATE=ON for these records.

# In[37]:


df.loc[
    case_3
    & (df["EXAMTM"].apply(to_secs) - df["PDMEDTM"].apply(to_secs) >= 1800)
    & (df["EXAMTM"].apply(to_secs) - df["PDMEDTM"].apply(to_secs) < 6 * 3600),
    "PDSTATE",
] = "ON"


# Updated records distribution:

# In[38]:


df.groupby(["PDSTATE", "PDTRTMNT"], dropna=False)[["REC_ID"]].count()


# In[39]:


case_3 = (df["PDSTATE"].isnull()) & (df["PDTRTMNT"] == 1)


# Number of records in case 3.b where PDMEDTM is earlier or equal to EXAMTM and $\text{EXAMTM-PDMEDTM} < 30\,min$:

# In[40]:


len(
    df[
        case_3
        & (df["EXAMTM"].apply(to_secs) - df["PDMEDTM"].apply(to_secs) < 1800)
        & (df["EXAMTM"].apply(to_secs) - df["PDMEDTM"].apply(to_secs) >= 0)
    ]
)


# Let's discard these records:

# In[41]:


df.drop(
    df[
        case_3
        & (df["EXAMTM"].apply(to_secs) - df["PDMEDTM"].apply(to_secs) < 1800)
        & (df["EXAMTM"].apply(to_secs) - df["PDMEDTM"].apply(to_secs) >= 0)
    ].index,
    inplace=True,
)


# Updated records distribution:

# In[42]:


df.groupby(["PDSTATE", "PDTRTMNT"], dropna=False)[["REC_ID"]].count()


# In[43]:


case_3 = (df["PDSTATE"].isnull()) & (df["PDTRTMNT"] == 1)


# Number of records in case 3.b where PDMEDTM is earlier or equal to EXAMTM and $\text{EXAMTM-PDMEDTM} \ge 6\,hours$:

# In[44]:


len(
    df[
        case_3
        & (df["EXAMTM"].apply(to_secs) - df["PDMEDTM"].apply(to_secs) >= 6 * 3600)
    ]
)


# Let's set PDSTATE=OFF for these records.

# In[45]:


df.loc[
    case_3 & (df["EXAMTM"].apply(to_secs) - df["PDMEDTM"].apply(to_secs) >= 6 * 3600),
    "PDSTATE",
] = "OFF"


# Updated records distribution:

# In[46]:


df.groupby(["PDSTATE", "PDTRTMNT"], dropna=False)[["REC_ID"]].count()


# ### Case 3.b.iii: PDMEDTM is later than EXAMTM

# Number of records in case 3.b where PDMEDTM is later than EXAMTM:

# In[47]:


case_3 = (df["PDSTATE"].isnull()) & (df["PDTRTMNT"] == 1)
len(df[case_3 & (df["PDMEDTM"] > df["EXAMTM"])])


# #### Case 3.b.iii.a: PDMEDTM is after 4pm

# Some of records have a PDMEDTIME which likely belongs to the previous day but there is no way to confirm it because dates only include a month, a year, but no day. Looking at the distribution of exam times in such cases, we noticed that no exam took place after 4pm, which motivates the use of 4pm as cut-off time.

# Distribution of EXAM times when $\text{PDMEDTM} > \text{EXAMTM}$:

# In[48]:


import numpy as np

pd.to_datetime(
    df[case_3 & (df["PDMEDTM"].apply(to_secs) > df["EXAMTM"].apply(to_secs))]["EXAMTM"]
).dt.hour.hist(bins=np.arange(24))
plt.xlabel("Hour of the day")
plt.ylabel("Number of EXAMTM records");


# Number of records in case 3 where PDMEDTM is after EXAMTM and PDMEDTM is after 4pm:

# In[49]:


len(
    df[
        case_3
        & (df["PDMEDTM"].apply(to_secs) > df["EXAMTM"].apply(to_secs))
        & (df["PDMEDTM"] > "16:00:00")
    ]
)


# Let's set PDSTATE=OFF for these records:

# In[50]:


df.loc[
    case_3
    & (df["PDMEDTM"].apply(to_secs) > df["EXAMTM"].apply(to_secs))
    & (df["PDMEDTM"] > "16:00:00"),
    "PDSTATE",
] = "OFF"


# Updated records distribution:

# In[51]:


df.groupby(["PDSTATE", "PDTRTMNT"], dropna=False)[["REC_ID"]].count()


# In[52]:


case_3 = (df["PDSTATE"].isnull()) & (df["PDTRTMNT"] == 1)


# #### Case 3.b.iii.a: PDMEDTM is before 4pm

# Number of records in case 3 where PDMEDTM is later than EXAMTM and PDMEDTM is before 4pm:

# In[53]:


len(df[case_3 & (df["EXAMTM"].apply(to_secs) < df["PDMEDTM"].apply(to_secs))])


# Let's remove these records. 

# In[54]:


df.drop(
    df[case_3 & (df["EXAMTM"].apply(to_secs) < df["PDMEDTM"].apply(to_secs))].index,
    inplace=True,
)


# In[55]:


case_3 = (df["PDSTATE"].isnull()) & (df["PDTRTMNT"] == 1)


# Let's verify that case 3 (PDSTATE=NaN, PDTRTMNT=1) is now resolved:

# In[56]:


df.groupby(["PDSTATE", "PDTRTMNT"], dropna=False)[["REC_ID"]].count()


# ## Case 4: PDSTATE=NaN and PDTRTMNT=NaN
# 
# Similar to case 1, no record in case 4 has a medication date (PDMEDDT), a medication time (PDMEDTM), or
# a DBS status (DBS_STATUS):

# In[57]:


case_4 = df[(df["PDSTATE"].isnull()) & (df["PDTRTMNT"].isnull())]
case_4.groupby(
    [
        "PDMEDDT",
        "PDMEDTM",
        "DBS_STATUS",
        "HRPOSTMED",
        "DBSONTM",
        "DBSOFFTM",
        "HRDBSOFF",
        "HRDBSON",
    ],
    dropna=False,
).count()


# <div class="alert alert-block alert-success">
#     	&#10003; <b>Proposed solution</b>: set PDSTATE=OFF and PDTRTMNT=0. It is very unlikely that the patient was medicated and none of the 8 medication-related variables were set.
# </div>

# ⚙️ Implementation

# In[58]:


df_1 = df.copy()
df_1.loc[(df["PDSTATE"].isnull()) & (df["PDTRTMNT"].isnull()), "PDSTATE"] = "OFF"
df_1.loc[(df["PDSTATE"].isnull()) & (df["PDTRTMNT"].isnull()), "PDTRTMNT"] = 0
df = df_1


# Let's verify that case 4 is now resolved:

# In[59]:


df.groupby(["PDSTATE", "PDTRTMNT"], dropna=False)[["REC_ID"]].count()


# ## Case 5: PDSTATE=ON and PDTRTMNT=NaN
# 
# 
# <div class="alert alert-block alert-success">
#     	&#10003; <b>Proposed correction</b>: set PDTRTMNT=1. The patient is medicated since PDSTATE=ON.
# </div>

# ⚙️ Implementation

# In[60]:


df.loc[(df["PDSTATE"] == "ON") & (df["PDTRTMNT"].isnull()), "PDTRTMNT"] = 1


# Let's verify that case 5 is now resolved:

# In[61]:


df.groupby(["PDSTATE", "PDTRTMNT"], dropna=False)[["REC_ID"]].count()


# There's no remaining missing PDSTATE or PDTRTMNT value in the data!
# 
# Let's save the cleaned file:

# In[62]:


filename = "MDS_UPDRS_Part_III_clean.csv"
df.to_csv(op.join(data_dir, filename), index=False)
print(f"Cleaned file saved in {filename}")


# # ⚕️ Sanity checks
# 
# The following sanity checks are applied to the cleaned UPDRS-III data.

# **IF** visit has two exam **THEN** one is ON and the other one is OFF:

# In[63]:


a = df.groupby(['PATNO', 'EVENT_ID']).filter(lambda x: len(x)==2)
a.groupby(['PATNO', 'EVENT_ID']).filter(lambda x: x.iloc[[0]]['PDSTATE'].to_string() ==
                                                  x.iloc[[1]]['PDSTATE'].to_string()).empty


# **IF** PDSTATE=ON **THEN** EXAMTM>PDMEDTM

# In[64]:


df[
    (df["PDSTATE"] == "ON")
    & (df["EXAMTM"].apply(to_secs) < df["PDMEDTM"].apply(to_secs))
].empty


# **IF** PDTRTMNT=0 **THEN** there is a single visit and PDSTATE=off

# In[65]:


assert (
    df[df["PDTRTMNT"] == 0]
    .groupby(["PATNO", "EVENT_ID"])
    .filter(lambda x: len(x) > 1)
    .empty
), "False!"
assert (
    df[df["PDTRTMNT"] == 0]
    .groupby(["PATNO", "EVENT_ID"])
    .filter(lambda x: x["PDSTATE"] != "OFF")
    .empty
), "False!"
print("True")


# A patient cannot become unmedicated after being medicated:

# In[66]:


def wrong_pairs(x):
    rows = [row for index, row in x.iterrows()]
    for a in rows:
        for b in rows:
            if a['EVENT_ID'] == b['EVENT_ID']:
                continue
            # If dates are equal, we cannot say anything
            if pd.to_datetime(a['INFODT']) == pd.to_datetime(b['INFODT']):
                return False                   
            # If a is later than b, a['PDTRTMNT'] has to be larger or equal to b['PDTRTMNT']
            if pd.to_datetime(a['INFODT']) > pd.to_datetime(b['INFODT']):
                if int(a['PDTRTMNT']) < int(b['PDTRTMNT']):
                    return True
                return False
            # a is earlier than b: a['PDTRTMNT'] has to be lower or equal to b['PDTRTMNT']
            if int(a['PDTRTMNT']) > int(b['PDTRTMNT']):
                return True
            return False

df.groupby(['PATNO']).filter(lambda x: x['PDTRTMNT'].nunique() > 1).groupby('PATNO').filter(wrong_pairs).empty


# In[67]:


# off = (df['PDSTATE']=='OFF') & (df['PDMEDTM']<='16:00:00')
# df['EXAMsecs'] = df[off]['EXAMTM'].apply(to_secs)
# df['PDMEDTMsecs'] = df[off]['PDMEDTM'].apply(to_secs)

# df[off & (df['EXAMsecs']-df['PDMEDTMsecs']>=0) & (df['EXAMsecs']-df['PDMEDTMsecs']<6*3600)]


# In[ ]:





# In[68]:


# off = (df['PDSTATE']=='OFF') & (df['PDMEDTM']<='16:00:00')
# df['EXAMsecs'] = df[off]['EXAMTM'].apply(to_secs)
# df['PDMEDTMsecs'] = df[off]['PDMEDTM'].apply(to_secs)

# df[off & (df['EXAMsecs']-df['PDMEDTMsecs']>=0) & (df['EXAMsecs']-df['PDMEDTMsecs']<6*3600)]


# In[ ]:







# noqa: D100
from setuptools import setup


install_requires = [
    "datalad",
    "IPython",
    "ppmi_downloader",
    "nilearn",
    "boutiques",
    "pillow",
    "matplotlib",
    "numpy",
    "pandas",
]

setup(
    name="livingpark_utils",
    version="0.4",
    description="Utility functions to write LivingPark notebooks.",
    author="Tristan Glatard",
    author_email="tristan.glatard@concordia.ca",
    license="MIT",
    packages=["livingpark_utils"],
    python_requires=">=3.10",
    install_requires=install_requires,
    include_package_data=True,
)



"""Provide utility function for the LivingPark notebook for paper replication."""
import datetime
import glob
import math
import os.path
import pkgutil
import subprocess
import sys
import warnings
from pprint import pprint

import nilearn.plotting as nplt
import numpy as np
import pandas as pd
import ppmi_downloader
import pytz  # type: ignore
from boutiques.descriptor2func import function as descriptor2func
from dateutil.parser import parse  # type: ignore
from dateutil.relativedelta import relativedelta  # type: ignore
from IPython.display import HTML
from IPython.display import Image as ImageDisplay
from matplotlib import pyplot as plt
from PIL import Image


class LivingParkUtils:
    """Contain functions to be reused across LivingPark notebooks."""

    def __init__(
        self,
        data_cache_path: str = ".cache",
    ) -> None:
        """Initialize a LivingPark notebook.

        Parameters
        ----------
        data_cache_path: str, default ".cache"
            Local path where to store the dataset cache.
            Keep default value unless you know what you're doing.
        """
        self.data_cache_path = data_cache_path
        self.study_files_dir = os.path.abspath(os.path.join("inputs", "study_files"))

        self.setup_notebook_cache()
        os.makedirs(self.study_files_dir, exist_ok=True)

    def setup_notebook_cache(self) -> None:
        """Create, install, and update the cache directory, if needed.

        Notes
        -----
        Aggregate the inputs and outputs into a single dataset by creating symlinks.
        """
        for x in ("", "inputs", "outputs"):
            os.makedirs(os.path.join(self.data_cache_path, x), exist_ok=True)

        # Make or update links to cache
        for x in ["inputs", "outputs"]:
            if os.path.islink(x):
                print(f"removing link {x}")
                os.remove(x)
            elif os.path.exists(x):
                raise Exception(f"Directory {x} exists and is not a symlink.")
            else:
                print(f"{x} doesnt exist")
            os.symlink(os.path.join(self.data_cache_path, x), x)

    def notebook_init(self) -> HTML:
        """Initialize a paper replication notebook.

        It ignores cell warnings, install dependencies, show execution time, and create
        a toggle button for displaying/hiding code cells.

        Returns
        -------
        HTML
            An HTML button to hide/show code cells in the notebooks.
        """
        warnings.filterwarnings("ignore")

        print("Installing notebook dependencies (see log in install.log)... ")
        f = open("install.log", "wb")
        subprocess.check_call(
            [sys.executable, "-m", "pip", "install", "-r", "requirements.txt"],
            stdout=f,
            stderr=f,
        )

        now = datetime.datetime.now(pytz.utc).strftime("%Y-%m-%d %H:%M:%S %Z %z")
        print(f"This notebook was run on {now}")

        return HTML(
            filename=os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                "toggle_button.html",
            )
        )

    def download_ppmi_metadata(
        self,
        required_files: list,
        force: bool = False,
        headless: bool = True,
        timeout: int = 600,
    ) -> None:
        """Download PPMI required study files, if not available.

        Parameters
        ----------
        required_files : list
            Required PPMI study files (cvs files) supported by ppmi_downloader.
        force : bool, optional
        headless : bool, default True
            If True, prevent broswer window to open during download.
        timeout : int, default 600
            Number of second before the download times out.

        Raises
        ------
        Exception:
            If failure occurs during download.
        """
        if force:
            missing_files = required_files
        else:
            missing_files = [
                x
                for x in required_files
                if not os.path.exists(os.path.join(self.study_files_dir, x))
            ]

        if len(missing_files) > 0:
            pprint(f"Downloading files: {missing_files}")
            try:
                ppmi = ppmi_downloader.PPMIDownloader()
                ppmi.download_metadata(
                    missing_files,
                    destination_dir=self.study_files_dir,
                    headless=headless,
                    timeout=timeout,
                )
            except Exception as e:
                print("Download failed!")
                raise (e)

            print("Download completed!")

        else:
            print("Download skipped: No missing files!")

    # def __install_datalad_cache(self) -> None:
    #     """Install the DataLad dataset.

    #     Notes
    #     -----
    #     Requires a functional ssh connection to `self.ssh_username`@`self.host`.
    #     Located at `self.host_dir`/`self.notebook_name`/`self.data_cache_path`.
    #     """
    #     if os.path.exists(self.data_cache_path):
    #         # noqa: TODO check if path is a valid DataLad dataset without doing d.status because it's too long.
    #         d = datalad.api.Dataset(self.data_cache_path)
    #         d.update(how="merge")
    #     else:
    #         datalad.api.install(
    #             source=(
    #                 f"{self.ssh_username}@{self.ssh_host}:"
    #                 f"{self.ssh_host_dir}/{self.notebook_name}"
    #             ),
    #             path=self.data_cache_path,
    #         )

    def clean_protocol_description(self, desc: str) -> str:
        """Create valid protocol description for file names (as done by PPMI).

        Parameters
        ----------
        str
            Protocol description. Example: "MPRAGE GRAPPA"
        """
        return (
            desc.replace(" ", "_").replace("(", "_").replace(")", "_").replace("/", "_")
        )

    def find_nifti_file_in_cache(
        self,
        subject_id: str,
        event_id: str,
        protocol_description: str,
        base_dir: str = "inputs",
    ) -> str | None:
        """Return cached nifti files, if any.

        Search for nifti file matching `subject_id`, `event_id` and
        `protocol_description` in the cache directory.
        If not found, search for nifti file matching `subject_id` and `event_id` only,
        and return it if a single file is found.

        Parameters
        ----------
        subject_id: str
            Subject ID
        event_id: str
            Event ID. Example: BL
        protocol_description: str
            Protocol description. Example: "MPRAGE GRAPPA"
        base_dir: str, default "inputs"
            TODO Describe this. Not sure what it is exactly.

        Returns
        -------
        str or None
            File name matching the `subject_id`, `event_id`, and if possible
            `protocol_description`. None if no matching file is found.
        """
        expression = os.path.join(
            self.data_cache_path,
            base_dir,
            f"sub-{subject_id}",
            f"ses-{event_id}",
            "anat",
            f"PPMI_*{self.clean_protocol_description(protocol_description)}*.nii",
        )
        files = glob.glob(expression)
        assert len(files) <= 1, f"More than 1 Nifti file matched by {expression}"
        if len(files) == 1:
            return files[0]
        # print(
        #     "Warning: no nifti file found for: "
        #     f"{(subject_id, event_id, protocol_description)} with strict glob "
        #     "expression. Trying with lenient glob expression."
        # )
        expression = os.path.join(
            self.data_cache_path,
            base_dir,
            f"sub-{subject_id}",
            f"ses-{event_id}",
            "anat",
            "PPMI_*.nii",
        )
        files = glob.glob(expression)
        assert len(files) <= 1, f"More than 1 Nifti file matched by {expression}"
        if len(files) == 1:
            return files[0]
        # print(
        #     f"Warning: no nifti file found for: "
        #     f"{(subject_id, event_id, protocol_description)} "
        #     "with lenient expression, returning None"
        # )
        return None

    def disease_duration(self) -> pd.DataFrame:
        """Return a DataFrame containing disease durations.

        Returns
        -------
        pd.DataFrame
            DataFrame containing disease durations for each (patient,event) pair found
            in "MDS_UPDRS_Part_III.csv".
        """
        # Download required files
        self.download_ppmi_metadata(
            ["MDS_UPDRS_Part_III.csv", "PD_Diagnosis_History.csv"]
        )

        pddxdt = pd.read_csv(
            os.path.join(self.study_files_dir, "PD_Diagnosis_History.csv")
        )[["PATNO", "EVENT_ID", "PDDXDT"]]
        pddxdt = pddxdt[(pddxdt["EVENT_ID"] == "SC") & pddxdt["PDDXDT"].notna()]
        pdxdur = pd.read_csv(
            os.path.join(self.study_files_dir, "MDS_UPDRS_Part_III.csv"),
            low_memory=False,
        )[["PATNO", "EVENT_ID", "INFODT"]]

        PDDXDT_map = dict(zip(pddxdt["PATNO"].values, pddxdt["PDDXDT"].values))
        pdxdur["PDDXDT"] = pdxdur["PATNO"].map(PDDXDT_map)

        pdxdur["PDXDUR"] = pdxdur.apply(
            lambda row: relativedelta(parse(row["INFODT"]), parse(row["PDDXDT"])).months
            if row["PDDXDT"] is not np.nan
            else np.nan,
            axis=1,
        )
        pdxdur.drop(labels=["INFODT", "PDDXDT"], inplace=True, axis=1)

        return pdxdur

    def moca2mmse(self, moca_score: int) -> int:
        """Return a MMSE score given a MoCA score.

        Parameters
        ----------
        moca_score: int
            MoCA score

        Returns
        -------
        int
            MMSE score corresponding to the MoCA score
            Conversion made using Table 2 in
            https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4371590

        """
        mapping = {
            1: 6,
            2: 9,
            3: 11,
            4: 12,
            5: 13,
            6: 14,
            7: 15,
            8: 15,
            9: 16,
            10: 17,
            11: 18,
            12: 18,
            13: 19,
            14: 20,
            15: 21,
            16: 22,
            17: 22,
            18: 23,
            19: 24,
            20: 25,
            21: 26,
            22: 26,
            23: 27,
            24: 28,
            25: 28,
            26: 29,
            27: 29,
            28: 30,
            29: 30,
            30: 30,
        }

        try:
            if math.isnan(moca_score):
                return np.nan
            else:
                return mapping[moca_score]
        except Exception as e:
            print(e)
            return moca_score

    def download_missing_nifti_files(
        self, cohort: pd.DataFrame, link_in_outputs=False
    ) -> None:
        """Download missing nifti files required by cohort.

        For each subject in cohort, look for T1-weighted nifti image file in
        notebook cache. Download all the missing files from PPMI, move them
        to notebook cache (inputs directory), and add their names to cohort.

        Parameters
        ----------
        cohort: pd.DataFrame
            A Pandas DataFrame containing columns PATNO (PPMI patient id), EVENT_ID
            (MRI visit, for instance 'V06'), and Description (for instance
            'MPRAGE GRAPPA'). Can be built from the file produced by
            'MRI metadata.ipynb'. A column 'File name' will be added to the DataFrame
             if not already present. This column
            will contain the paths of the T1-weighted nifti files associated with the
             patient, MRI visit, and protocol description.

        link_in_outputs: bool
            If True, create symbolic links to input nifti files in
            outputs/pre-processing. Useful for processing tools that
            write next to input files, such as SPM.

        Returns
        -------
        None
        """
        # Find nifti file names in cohort
        cohort["File name"] = cohort.apply(
            lambda x: self.find_nifti_file_in_cache(
                x["PATNO"], x["EVENT_ID"], x["Description"]
            ),
            axis=1,
        )
        print(
            f"Number of available subjects: {len(cohort[cohort['File name'].notna()])}"
        )
        print(f"Number of missing subjects: {len(cohort[cohort['File name'].isna()])}")

        # Download missing file names
        try:
            ppmi_dl = ppmi_downloader.PPMIDownloader()
            missing_subject_ids = cohort[cohort["File name"].isna()]["PATNO"]
            print(f"Downloading image data of {len(missing_subject_ids)} subjects")
            ppmi_dl.download_imaging_data(
                missing_subject_ids,
                type="nifti",
                timeout=120 * len(missing_subject_ids),
                headless=False,
            )
        except Exception as e:
            print("Download failed!")
            raise (e)

        # Find cohort file names among downloaded files
        results_path = "outputs"
        ppmi_fd = ppmi_downloader.PPMINiftiFileFinder()
        for _, row in cohort.iterrows():
            if row["File name"] is None:
                filename = ppmi_fd.find_nifti(
                    row["PATNO"], row["EVENT_ID"], row["Description"]
                )
                if filename is None:
                    print(
                        "Not found: "
                        + f"{row['PATNO'], row['EVENT_ID'], row['Description']}"
                    )
                else:  # copy file to dataset
                    dest_dir = os.path.join(
                        "inputs",
                        f'sub-{row["PATNO"]}',
                        f'ses-{row["EVENT_ID"]}',
                        "anat",
                    )
                    os.makedirs(dest_dir, exist_ok=True)
                    dest_file = os.path.join(dest_dir, os.path.basename(filename))
                    os.rename(filename, dest_file)
                    row["File name"] = dest_file

        # Update file names in cohort
        cohort["File name"] = cohort.apply(
            lambda x: self.find_nifti_file_in_cache(
                x["PATNO"], x["EVENT_ID"], x["Description"]
            ),
            axis=1,
        )

        # Create symlinks to inputs if necessary
        if link_in_outputs:
            for file_name in cohort["File name"].values:
                dest_dir = os.path.dirname(file_name).replace(
                    os.path.join(self.data_cache_path, "inputs"),
                    os.path.join(results_path, "pre_processing"),
                )
                dest_file = os.path.join(
                    dest_dir,
                    os.path.basename(file_name.replace(self.data_cache_path, "")),
                )
                if not os.path.exists(dest_file):
                    # print(dest_dir, file_name, dest_file)
                    os.makedirs(dest_dir, exist_ok=True)
                    os.symlink(
                        os.path.relpath(os.path.abspath(file_name), start=dest_file),
                        dest_file,
                    )

    def cohort_id(self, cohort: pd.DataFrame) -> str:
        """Return a unique id for the cohort.

        The id is built as the hash of the sorted list of patient ids in the cohort.
        Since cohort_ids may be used to create file names, negative signs ('-')
        are replaced with underscore characters ('_') since SPM crashes on file names
        containing negative signs. Therefore, the cohort id is a string that cannot
        be cast to an integer.

        Parameters
        ----------
        cohort: pd.DataFrame
            A Pandas DataFrame with a column named 'PATNO'.

        Return
        ------
        string
            A string containing the unique id of the cohort.
        """
        return str(hash(tuple(sorted(cohort["PATNO"])))).replace("-", "_")

    def write_spm_batch_files(
        self,
        template_job_filename: str,
        replaced_keys: dict,
        executable_job_file_name: str,
    ) -> None:
        """Write SPM batch files from a template by replacing placeholder keys in it.

        Open the SPM batch file in template_job_filename, search and replace keys found
        in replaced_keys, and write the result in two files, a "batch" file and a "job"
        file. Output file names are built from job_file_name. Job file names must end
        with '_job.m'

        Parameters
        ----------
        template_job_filename: str
            File name of template SPM job. Contains placeholder keys to be replaced to
            create an executable batch. No format is specified for the keys, make sure
            that they are uniquely identified in the template file!

        replaced_keys: dict
            Dictionary containing keys to be replaced by values in the template job
            file. Example: {'[IMAGE]': 'inputs/sub-1234/ses-1/anat/image.nii'}. Make
            sure that the keys are present in the template job file name!

        executable_job_file_name: str
            File name where to write the executable job file. Must end in '_job.m'.
            An SPM batch file calling this job file will also be written with a
            '_batch.m' ending.

        Return
        ------
        None
        """

        def replace_keys(string, replace_keys):
            for k in replace_keys:
                string = string.replace(k, replace_keys[k])
            return string

        # Read template file
        with open(template_job_filename) as f:
            content = f.read()

        assert template_job_filename.endswith("_job.m")
        assert executable_job_file_name.endswith("_job.m")

        with open(executable_job_file_name, "w") as f:
            f.write(replace_keys(content, replaced_keys))

        print(f"Job batch file written in {os.path.basename(executable_job_file_name)}")

        # Batch file
        content_batch = pkgutil.get_data(
            __name__, os.path.join("templates", "call_batch.m")
        )
        assert content_batch is not None, "Cannot read batch template file."
        content_batch_str = content_batch.decode("utf-8")
        tempfile_name_batch = executable_job_file_name.replace("_job", "_batch")

        with open(tempfile_name_batch, "w") as f:
            job_dir = os.path.dirname(os.path.abspath(executable_job_file_name))
            f.write(
                replace_keys(
                    content_batch_str,
                    {
                        "[BATCH]": f"addpath('{job_dir}')"
                        + os.linesep
                        + os.path.basename(executable_job_file_name.replace(".m", ""))
                    },
                )
            )

        print(f"Batch file written in {os.path.basename(tempfile_name_batch)}")

    def run_spm_batch_file(
        self,
        executable_job_file_name: str,
        boutiques_descriptor: str = "zenodo.6881412",
        force: bool = False,
    ) -> None:
        """Run an SPM batch file using Boutiques.

        Requires Docker or Singularity container engines (Singularity untested yet in
        this context). Download the Boutiques descriptor from Zenodo or use the local
        file passed as argument. Download the Docker container, create a Boutiques
        invocation and run it. Write logs in log file created from
        executable_job_file_name (example: pre_processing_1234.log). If log file
        already exists, skip execution unless force is set to True.

        Parameters
        ----------
        executable_job_file_name: str
            An SPM job file ready to be executed.
            Example: 'code/batches/pre_processing_1234_job.m'.
            See self.write_spm_batch_files for a possible way to create such a file.


        boutiques_descriptor: str
            A Boutiques descriptor in the form of a Zenodo id, local file name, or
            JSON string. Don't modify the default value unless you know what you are
            doing.

        force: bool
            Force execution even if log file already exists for this execution.
            Default: False.

        Return
        ------
        boutiques.ExecutionOutput
            Boutiques execution output object containing exit code and various logs.
        """
        log_dir = os.path.join("outputs", "logs")
        os.makedirs(log_dir, exist_ok=True)

        log_file_name = os.path.abspath(
            os.path.join(
                log_dir,
                os.path.basename(executable_job_file_name.replace("_job.m", ".log")),
            )
        )
        spm_batch_file = executable_job_file_name.replace("_job", "_batch")

        if not force:
            if os.path.exists(log_file_name):
                print(
                    f"Log file {os.path.basename(log_file_name)} exists, "
                    + "skipping batch execution (remove file or use force=True "
                    + "to force execution)"
                )
                return
            else:
                print(
                    f"Log file {os.path.basename(log_file_name)} does not exist, "
                    + "running batch"
                )

        # Initialize Boutiques Python function for descriptor
        spm_batch = descriptor2func(boutiques_descriptor)

        output = spm_batch(
            "launch",
            "-s",
            "-u",
            spm_batch_file=spm_batch_file,
            log_file_name=log_file_name,
        )

        assert (
            output.exit_code == 0
        ), f"Execution error, inspect output object for logs: {output}"

        print("Execution was successful.")

        return output

    def smwc_scan(
        self,
        tissue_class: int,
        patno: int,
        visit: str,
        pre_processing_dir: str = "pre_processing",
    ) -> str:
        """Find the SPM tissue class file of patient at visit with given protocol.

        Scans the outputs directory for an SPM tissue class file obtained from nifti
        file of patient at visit using protocol description. To find a tissue file,
        matches glob expression
        outputs/{pre_processing_dir}sub-{patno}/ses-{visit}/anat/smwc{tissue_class}PPMI*.nii
        Returns an error if more than one file is found that matches this expression.

        Paramters
        ---------
        tissue_class: int
            1 (grey matter) or 2 (white matter)

        patno: int
            PPMI patient identifier

        visit: str
            PPMI visit name. Example: 'V04'.

        pre_processing_dir: str
            Directory in 'outputs' where pre-processing results are stored.

        Return
        ------
        str: path of a tissue class file
        """
        if tissue_class not in (1, 2):
            raise Exception(f"Unrecognized tissue class: {tissue_class}")
        dirname = os.path.join("outputs", pre_processing_dir)
        expression = (
            f"{dirname}/sub-{patno}/ses-{visit}/anat/smwc{tissue_class}PPMI*.nii"
        )
        files = glob.glob(expression)
        assert (
            len(files) == 1
        ), f"Zero or more than 1 files were matched by expression: {expression}"
        return os.path.abspath(files[0])

    def export_spm_segmentations(
        self,
        cohort: pd.DataFrame,
        folder: str,
        cut_coords: tuple = (-28, -7, 17),
        force: bool = False,
        extension: str = "png",
    ) -> None:
        """Export segmentation images as 2D image files.

        Meant to be used for quality control. make_gifs can assemble these images
        in an animate gif.

        Parameters
        ----------
        cohort: pd.DataFrame
            LivingPark cohort to export. Must have a column called 'File name'.

        folder: str
            Folder where to export the segmentation images.

        cut_coords: tuple
            Passed to Nilearn viewer. The MNI coordinates of the cutting plane.

        force: bool
            If True, force export to existing folder. Removes all the files in folder
            before writing new ones.

        extension: str
            Image file extension supported by Matplotlib. Example: 'png'.
        """
        alpha = 0.5

        if (not force) and os.path.exists(folder):
            print(
                f"Folder {folder} already exists, skipping image export "
                + " (remove folder or use --force to force)."
            )
            return

        if os.path.exists(folder):  # force is True
            print(f"Folder {folder} already exists, removing its content")
            for f in glob.glob(os.path.join(folder, "*")):
                os.remove(f)

        os.makedirs(folder, exist_ok=True)

        for i in range(len(cohort)):

            input_file = cohort["File name"].values[i]
            subj_id = cohort["PATNO"].values[i]

            output_file_name = input_file.replace(
                os.path.join(self.data_cache_path, "inputs"),
                os.path.join("outputs", "pre_processing"),
            )
            output_file_c1 = output_file_name.replace("PPMI", "smwc1PPMI")
            output_file_c2 = output_file_name.replace("PPMI", "smwc2PPMI")

            fig = plt.figure()
            display = nplt.plot_anat(
                cut_coords=list(cut_coords), figure=fig, title=f"#{i}/{len(cohort)}"
            )
            display.add_overlay(output_file_c1, cmap="Reds", threshold=0.1, alpha=alpha)
            display.add_overlay(
                output_file_c2, cmap="Blues", threshold=0.1, alpha=alpha
            )

            os.makedirs(folder, exist_ok=True)
            plt.savefig(
                os.path.join(
                    folder, f"qc_{self.cohort_id(cohort)}_{subj_id}.{extension}"
                )
            )
            plt.close(fig)  # so as to not display the figure

    def make_gif(self, frame_folder: str, output_name: str = "animation.gif") -> None:
        """Make gifs from a set of images located in the same folder.

        Parameters
        ----------
        frame_folder : str
            Folder where frames are stored. Frames must be in a format supported by PIL.

        output_name : str
            Base name of the gif file. Will be written in frame_folder.
        """
        frames = [
            Image.open(image)
            for image in glob.glob(os.path.join(f"{frame_folder}", "*.png"))
        ]
        frame_one = frames[0]
        frame_one.save(
            os.path.join(frame_folder, output_name),
            format="GIF",
            append_images=frames,
            save_all=True,
            duration=1000,
            loop=0,
        )
        print(f"Wrote {os.path.join(frame_folder, output_name)}")

    def qc_spm_segmentations(self, cohort) -> None:
        """Display a gif file with SPM segmentation results from the cohort.

        Parameters
        ----------
        cohort: pd.DataFrame
            LivingPark cohort to QC. Must have a column called 'File name'.
        """
        qc_dir = f"qc_{self.cohort_id(cohort)}"

        self.export_spm_segmentations(cohort, qc_dir)
        self.make_gif(qc_dir)
        return ImageDisplay(url=os.path.join(qc_dir, "animation.gif"))
